!---------------------------------------------------------------------------------------
! PREP_REGIONS: map_grid
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> map_grid
!!
!! Purpose:    Calculates the mapping to an aggregated grid based on the flux
!!             sensitivities from FLEXPART
!!
!! Details:    Based on mapgrid.f routine of A. Stohl. This subroutine lumps grid cells 
!!             with low flux sensitivity into larger boxes. This is done in nmap steps, 
!!             starting with the coarsest boxes (resolution nres set by mapdec). 
!!             When both the sensitivity is among the "cutoff" lowest values, boxes are 
!!             used at coarse resolution. If they have higher sensitivity values, they 
!!             are taken to the next step (i.e., boxes are split to higher resolution). 
!!             This routine differs w.r.t old FLEXINVERT routine in that boxes that 
!!             are >99% water are given negative numbers while boxes that are land are
!!             given positive numbers.
!!
!! Options (defined in SETTINGS): 
!!             usezoom         -  (true/false) global map with fine resolution only over region 
!!                                 of interest (i.e. nested region in SETTINGS_config)
!!             useflux         -  (true/false) weight surface influence by annual fluxes
!!             uselanduse      -  (true/false) avoid gridcells with > 99% water
!!             nmap            -  (integer) number of resolution steps to use
!!             cutoff          -  (real) upper fractile of the sensitivity used to define threshold
!!             mapdec          -  (integer vector of length nmap) defines resolution steps
!!                                 e.g. if flexpart resolution is 0.5 degree then mapdec = (/1,2,2/)
!!                                 will give 3 resolution steps: 
!!                                 1) 2 x 2 x 1 x 0.5 = 2 degrees
!!                                 2) 2 x 1 x 0.5 = 1 degree
!!                                 3) 1 x 0.5 = 0.5 degree 
!!
!! Interface:
!!
!!    Inputs
!!             config   -  configuration settings data structure
!!             flux     -  fluxes
!!             surfinf  -  total flux sensitivity
!!             nbox_xy  -  mapping matrix (zero matrix on input)
!!
!!    Externals
!!             sort
!!
!---------------------------------------------------------------------------------------

subroutine map_grid(config, flux, surfinf, nbox_xy)

  use mod_var
  use mod_settings

  implicit none

  type (config_t),                    intent(in)     :: config
  real, dimension(nxregrid,nyregrid), intent(in)     :: flux
  real, dimension(nxregrid,nyregrid), intent(in)     :: surfinf
  real, dimension(nxregrid,nyregrid), intent(in out) :: nbox_xy

  real, parameter                                    :: sensitivity_min = 1.e15
  integer, parameter                                 :: maxbox = 50000
  real                                               :: flux_av
  integer                                            :: n, i, ix, jy, iix, jjy, i1, j1, ii, jj
  integer                                            :: numx, numy, num
  integer                                            :: nres, ndec, nfx, nfy, nfbox, nsort
  integer, dimension(2*maxbox)                       :: map_box
  integer                                            :: nbox_invert, nb, nbox_land, nbox_water
  real, dimension(2*maxbox)                          :: x_map, y_map, sens_map, water_box
  real, dimension(nxregrid,nyregrid)                 :: sens_lump, contrib_lump, water
  real, dimension(nxregrid*nyregrid)                 :: sort_contrib, sort_contrib_inv
  logical, dimension(nxregrid,nyregrid)              :: mark 
  logical                                            :: zoomflag
  real, dimension(nxregrid)                          :: lon_lump
  real, dimension(nyregrid)                          :: lat_lump
  real                                               :: cut_contrib

  ! initialize some variables
  nbox_invert = 0
  nbox_xy(:,:) = 0
  sens_lump(:,:) = 0.
  contrib_lump(:,:) = 0.
  sort_contrib(:) = 0.
  water_box(:) = 0.
  water(:,:) = 0.
  lon_lump(:) = 0.
  lat_lump(:) = 0.
  zoomflag = .false.
  ! mean flux 
  flux_av = sum(flux)/float(nxregrid*nyregrid)

  ! increase resolution stepwise from coarsest to finest grid
  ! first iteration at resolution equal to product of mapdec
  do ndec = nmap, 2, -1
    nres = 1
    do i = 1, ndec
      nres = nres*mapdec(i)
    end do

    ! lump data on coarse grid
    nfx = 0
    numx = (nxregrid/nres)*nres
    if (numx.lt.nxregrid) then 
      numx = numx + nres
    endif 
    do ix = 1, numx, nres
      nfx = nfx + 1
      nfy = 0
      lon_lump(nfx) = rllx + real((nfx-1)*nres)
      numy = (nyregrid/nres)*nres
      if (numy.lt.nyregrid) then              
        numy = numy + nres
      endif
      do jy = 1, numy, nres
        nfy = nfy + 1
        lat_lump(nfy) = rlly + real((nfy-1)*nres)
        sens_lump(nfx,nfy) = 0.
        contrib_lump(nfx,nfy) = 0.
        water(nfx,nfy) = 0.
        mark(nfx,nfy) = .false.
        ! sens_lump : contains sum of surface influences from all grid cells at 
        ! fine resolution corresponding to the grid cell at coarse resolution 
        num = 0
        do i1 = 0, nres-1
          do j1 = 0, nres-1
            if ( (ix+i1).gt.nxregrid ) go to 10
            if ( (jy+j1).gt.nyregrid ) go to 10
            num = num + 1
            if ( nbox_xy(ix+i1,jy+j1).gt.0 ) then
              mark(nfx,nfy) = .true.
            else
              sens_lump(nfx,nfy) = sens_lump(nfx,nfy) + surfinf(ix+i1,jy+j1)
              if (useflux) then
                ! contrib_lump : surface influence times flux 
                ! take a certain minimum flux flux for sorting
                contrib_lump(nfx,nfy) = contrib_lump(nfx,nfy) + surfinf(ix+i1,jy+j1) * abs(flux(ix+i1,jy+j1))/flux_av
              else
                ! contrib_lump : surface influence only
                contrib_lump(nfx,nfy) = contrib_lump(nfx,nfy) + surfinf(ix+i1,jy+j1)
              endif
              if( uselanduse ) then
                water(nfx,nfy) = water(nfx,nfy) + (1. - lsm(ix+i1,jy+j1))
              else
                ! set to zero to avoid landuse consideration
                water(nfx,nfy) = 0.
              endif
            endif
10          continue
          end do ! nres
        end do ! nres
        water(nfx,nfy) = water(nfx,nfy)/float(num)
      end do ! nyregrid
    end do ! nxregrid
    nfbox = nfx*nfy

    ! sort values in decreasing order and determine cut-off 
    nsort = 0
    do iix = 1, nfx
      do jjy = 1, nfy
        if (.not.mark(iix,jjy)) then
          nsort = nsort + 1
          sort_contrib(nsort) = contrib_lump(iix,jjy)
        endif
      end do ! nfy
    end do ! nfx
    ! sort: orders in increasing order
    call sort(nsort,sort_contrib(1:nsort))
    ! change to decreasing order
    sort_contrib_inv(:) = 0.
    do n = 1, nsort
      sort_contrib_inv(nsort-n+1) = sort_contrib(n)
    end do
    sort_contrib = sort_contrib_inv
    cut_contrib = sort_contrib(nint(cutoff*float(nsort))+1)

    print*, 'lon_lump: ',lon_lump
    print*, 'lat_lump: ',lat_lump

    ! for all lumped boxes below threshold, mark lumped box, which will
    ! subsequently not be subdivided any further
    do iix = 1, nfx
      do jjy = 1, nfy
        if ( usezoom ) then
          if ( lon_lump(iix).lt.zllx.or.(lon_lump(iix)+real(nres)).gt.zurx.or. &
              lat_lump(jjy).lt.zlly.or.(lat_lump(jjy)+real(nres)).gt.zury ) then 
            zoomflag = .true.
          else
            zoomflag = .false.
          endif
        endif
        if(nres==8) print*, 'zoomflag = ',zoomflag
        if ( (.not.mark(iix,jjy)).and.((contrib_lump(iix,jjy).le.cut_contrib).or.zoomflag.or.(water(iix,jjy).gt.0.99)) ) then
          if ( zoomflag.or.(water(iix,jjy).gt.0.99).or.(sens_lump(iix,jjy).lt.sensitivity_min) ) then
!        if ((.not.mark(iix,jjy)).and.((contrib_lump(iix,jjy).le.cut_contrib)).or.(water(iix,jjy).gt.0.99)) then
!          if ((water(iix,jjy).gt.0.99).or.(sens_lump(iix,jjy).lt.sensitivity_min)) then
            nbox_invert = nbox_invert + 1
            if (nbox_invert.gt.maxbox) then
              write(*,*) 'ERROR: maxbox too small'
              write(*,*) 'nbox_invert = ',nbox_invert
              stop
            endif
            nb = nbox_invert
            map_box(nb) = ndec
            x_map(nb) = iix
            y_map(nb) = jjy
            sens_map(nb) = sens_lump(iix,jjy)
            water_box(nb) = water(iix,jjy)
            do ix = (iix-1)*nres+1,iix*nres
              do jy = (jjy-1)*nres+1,jjy*nres
                ii = min(nxregrid,ix)
                jj = min(nyregrid,jy)
                nbox_xy(ii,jj) = nb
              end do
            end do
          endif 
        endif
      end do ! nfy
    end do ! nfx

  end do ! loop over nmap

  ! for all remaining boxes, use the full-resolution model grid
  ndec = 1
  do ix = 1, nxregrid
    do jy = 1, nyregrid
      if (nbox_xy(ix,jy).eq.0) then
        ! use box for matrix inversion
        nbox_invert = nbox_invert + 1
        if (nbox_invert.gt.maxbox) then
          write(*,*) 'ERROR: maxbox too small'
          write(*,*) 'nbox_invert = ',nbox_invert
          stop
        endif    
        nb = nbox_invert
        map_box(nb) = ndec
        x_map(nb) = ix
        y_map(nb) = jy
        sens_map(nb) = surfinf(ix,jy)
        water_box(nb) = 1. - lsm(ix,jy)
        nbox_xy(ix,jy) = nb
      endif
    end do ! nxregrid
  end do ! nyregrid

  print*, 'nbox_invert, nb = ',nbox_invert, nb

  ! renumber boxes for land (1:nbox_land) and water (-1:-nbox_water) 
  nbox_water = 0
  nbox_land = 0
  do nb = 1, nbox_invert
    if ( water_box(nb).gt.0.99 ) then 
      ! water
      nbox_water = nbox_water - 1
      where( nbox_xy.eq.nb ) nbox_xy = nbox_water
    else
      ! land
      nbox_land = nbox_land + 1
      where( nbox_xy.eq.nb ) nbox_xy = nbox_land
    endif
  end do

end subroutine map_grid

